#!/usr/bin/env python3

import os
import pprint
import subprocess
import sys

from charmhelpers.core import hookenv

from lib.ubuntu_repository_cache import (
    metadata_sync,
    mirror,
    service,
    util,
)

LOG = hookenv.log
SERVICE = 'ubuntu-repository-cache'
PRETTYPRINT = pprint.PrettyPrinter(indent=4)

HOOKS = hookenv.Hooks()


@HOOKS.hook('debug')
def debug():
    '''Debug hook for development'''
    PRETTYPRINT.pprint(hookenv.execution_environment())


def parse_config(conf=service.CONFIG_PATH):
    keys = {}
    cmd = ['bash', '-c', "trap 'env -0' exit; source {} > /dev/null 2>&1".format(conf)]
    output = subprocess.Popen(cmd, shell=False, stdout=subprocess.PIPE).communicate()[0]
    for ll in output.decode('utf-8').split('\x00'):
        if not ll:
            continue
        key, value = ll.split('=', 1)
        keys[key] = value

    return keys


@HOOKS.hook('ubuntu_repository_cache_sync')
@util.run_as_user('www-sync')
def ubuntu_repository_cache_sync():
    """Notify clustered peers that metadata is available.

    An external system (a cron job) will trigger this hook after it has
    finished syncing metadata from upstream. This will notify the peers that
    metadata is ready."""

    LOG('Notifying peers of metadata update')
    # We can't use hookenv.config() here because the persistent config
    # permissions are restricted to root:root 0600 - LP:1828658
    apache_root = parse_config()['APACHE_ROOT']
    apache_data = '/'.join((apache_root, 'data'))

    meta_ver = sys.argv[1]
    if not meta_ver:
        LOG('Leader could not gather new metadata', hookenv.WARNING)
        return

    link_next = '/'.join((apache_data, 'ubuntu_next'))

    meta_dir = os.path.realpath(link_next)

    # Push the metadata to peers
    if hookenv.is_relation_made('cluster'):
        dest = '/'.join((apache_data, meta_ver))
        link_dir = '/'.join((apache_data, 'ubuntu_active'))
        LOG('Pushing updates to peers', hookenv.DEBUG)
        source = meta_dir + '/'
        successful_peers = mirror.rsync_to_peers(source, dest, link_dir=link_dir)

        # Push metaversion_good canary to indicate a complete sync
        canary = '_'.join((meta_dir, 'good'))
        LOG('Pushing canary files to peers', hookenv.DEBUG)
        mirror.rsync_to_peers(canary, apache_data, peers=successful_peers)

        # Update 'meta_version' relation setting
        LOG('', hookenv.DEBUG)
        ids = hookenv.relation_ids('cluster') or []
        LOG('Updating meta_version for ids %s' % ids, hookenv.DEBUG)
        for r_id in ids:
            LOG('Setting meta_version=%s' % meta_ver, hookenv.DEBUG)
            hookenv.relation_set(relation_id=r_id, meta_version=meta_ver)

    util.update_active_symlink(os.path.basename(meta_dir), apache_root)
    metadata_sync.clean_metadata_dir(apache_root, LOG)
    service.start()


if __name__ == "__main__":
    # execute a hook based on the name the program is called by
    HOOKS.execute(sys.argv)
