'''Functions related to apache use within this charm'''

import multiprocessing
import os
import subprocess
import json

from charmhelpers import (
    fetch,
)

from charmhelpers.core import (
    hookenv,
    host,
    templating,
    unitdata,
)
from . import util


LOG = hookenv.log
SERVICE = 'apache2'


@util.run_as_user('root')
def install():
    '''Perform the install/config of apache and disable the default site'''

    LOG('Installing apache', hookenv.INFO)
    fetch.apt_install(SERVICE, fatal=True)
    stop()
    subprocess.check_call(['a2enmod', 'expires'])
    subprocess.check_call(['a2enmod', 'proxy'])
    subprocess.check_call(['a2enmod', 'proxy_http'])
    subprocess.check_call(['a2enmod', 'headers'])
    subprocess.check_call(['a2dismod', 'mpm_event'])
    subprocess.check_call(['a2enmod', 'mpm_worker'])
    subprocess.check_call(['a2dissite', '000-default'])


@util.run_as_user('root')
def start():
    '''Start the service and remove the override file blocking startup'''

    override_path = '/etc/init/apache2.override'
    if os.path.exists(override_path):
        os.remove('/etc/init/apache2.override')

    if service_running():
        LOG('Reloading apache', hookenv.INFO)
        host.service_reload(SERVICE)
    else:
        LOG('Starting apache', hookenv.INFO)
        host.service_start(SERVICE)


@util.run_as_user('root')
def stop():
    '''Stop the service and block startup at boot'''

    LOG('Stopping apache', hookenv.INFO)
    with open('/etc/init/apache2.override', 'w+') as override_file:
        override_file.write('manual\n')

    host.service_stop(SERVICE)


@util.run_as_user('root')
def service_running():
    '''Return True is the service is running'''

    return host.service_running(SERVICE)


def tune_mpm_configs(config):
    cpu_count = multiprocessing.cpu_count()

    if not config['startservers']:
        config['startservers'] = 1

    if not config['serverlimit']:
        # Benchmarking shows single process multiple threads performs better
        # than multiple processes multiple threads.
        config['serverlimit'] = 1

    if not config['threadsperchild']:
        # We could take into consideration how much memory the system
        # has but the ratio of vCPU to RAM between public clouds seems
        # good enough so this is not needed.
        config['threadsperchild'] = 256 * cpu_count
        # Cap at 8192
        if config['threadsperchild'] > 8192:
            config['threadsperchild'] = 8192

    if not config['threadlimit']:
        # Hard limit should match per child limit.
        config['threadlimit'] = config['threadsperchild']

    if not config['maxrequestworkers']:
        # Calculate max. workers based on serverlimit and threadsperchild.
        config['maxrequestworkers'] = config['serverlimit'] * config['threadsperchild']

    if not config['minsparethreads']:
        # Use 50% of max. workers.
        config['minsparethreads'] = int(config['maxrequestworkers'] * 0.5)

    if not config['maxsparethreads']:
        config['maxsparethreads'] = config['maxrequestworkers']

    return config


def create_mpm_workerfile():
    '''Create the multi-processing module (MPM) configuration'''

    config = hookenv.config()
    mpm_context = {
        'startservers': config['apache2_mpm_startservers'],
        'minsparethreads': config['apache2_mpm_minsparethreads'],
        'maxsparethreads': config['apache2_mpm_maxsparethreads'],
        'threadlimit': config['apache2_mpm_threadlimit'],
        'threadsperchild': config['apache2_mpm_threadsperchild'],
        'serverlimit': config['apache2_mpm_serverlimit'],
        'maxrequestworkers': config['apache2_mpm_maxrequestworkers'],
        'maxconnectionsperchild': config['apache2_mpm_maxconnectionsperchild'],
    }
    mpm_context = tune_mpm_configs(mpm_context)
    conf_mpm_worker = '/etc/apache2/conf-available/000mpm-worker.conf'
    templating.render('apache2/mpm_worker.template', conf_mpm_worker, mpm_context)
    subprocess.check_call(['a2enconf', '000mpm-worker'])
    mods_mpm_worker = '/etc/apache2/mods-available/mpm_worker.conf'
    if not os.path.exists(mods_mpm_worker):
        os.symlink(conf_mpm_worker, mods_mpm_worker)
    elif not os.path.islink(mods_mpm_worker):
        os.unlink(mods_mpm_worker)
        os.symlink(conf_mpm_worker, mods_mpm_worker)


def create_security():
    '''Create the security configuration'''

    config = hookenv.config()
    security_context = {
        'server_tokens': config['apache2_server_tokens'],
        'server_signature': config['apache2_server_signature'],
        'trace_enabled': config['apache2_trace_enabled'],
    }
    templating.render('apache2/security.template', '/etc/apache2/conf-available/security.conf', security_context)
    subprocess.check_call(['a2enconf', 'security'])


def create_metadata_site():
    '''Create the site config for serving (or proxying) repository metadata'''

    config = hookenv.config()
    apache_context = {}
    apache_context['DocumentRoot'] = unitdata.kv().get('apache-root')
    apache_context['Sync_Host'] = config['sync-host']
    apache_context['Display_Host'] = config['display-host']
    apache_context['Path_Base'] = config['path-base']
    metaproxy_conf = '# No failover configured'
    failover_host = util.get_failover_host()
    if util.in_failover() and failover_host:
        uri = 'http://%s:80' % failover_host
        metaproxy_conf = (
            'ProxyRemote      http://{disp}/ {uri}\n'
            '\tProxyPass        / http://{disp}/\n'
            '\tProxyPassReverse / http://{disp}/'
        ).format(uri=uri, disp=config['display-host'])
    apache_context['MetadataProxy'] = metaproxy_conf

    templating.render(
        'apache2/archive_ubuntu_com.conf', '/etc/apache2/sites-available/archive_ubuntu_com.conf', apache_context
    )
    subprocess.call(['a2ensite', 'archive_ubuntu_com'])
    start()


@util.run_as_user('root')
def configure_remoteip_logging():
    config = hookenv.config()
    if config.changed('remoteip_logging'):
        LOG('Apache remoteip_logging changed')
        if config['remoteip_logging'] is True:
            subprocess.check_call(['a2enmod', 'remoteip'])
            templating.render('apache2/remoteip.conf', '/etc/apache2/conf-available/remoteip.conf', {})
            subprocess.check_call(['a2enconf', 'remoteip'])
            LOG('Enabled remoteip logging')
        else:
            subprocess.check_call(['a2dismod', 'remoteip'])
            if os.path.exists('/etc/apache2/conf-enabled/remoteip.conf'):
                subprocess.check_call(['a2disconf', 'remoteip'])
            LOG('Disabled remoteip logging')

        start()


@util.run_as_user('root')
def render_configs():
    '''Render the configuration templates for apache'''

    LOG('Rendering apache2 configuration templates')
    config = hookenv.config()
    if config.changed('apache2_mpm_type'):
        if config['apache2_mpm_type'] == 'worker':
            LOG('Using apache2-mpm-worker')
            subprocess.check_call(['a2dismod', 'mpm_prefork'])
            subprocess.check_call(['a2enmod', 'mpm_worker'])
        elif config['apache2_mpm_type'] == 'prefork':
            LOG('Using apache2-mpm-prefork')
            # Package only in Trusty and below.
            subprocess.check_call(['a2dismod', 'mpm_worker'])
            subprocess.check_call(['a2enmod', 'mpm_prefork'])

    create_metadata_site()
    create_mpm_workerfile()
    create_security()
    configure_remoteip_logging()

    start()


def update_checks(nrpe_config):
    '''Update nagios check for apache serving archive metadata'''

    path_base = hookenv.config()['path-base']
    try:
        with open('/run/cloud-init/status.json') as status:
            status_json = json.load(status)
    except OSError:
        pass
    else:
        # try to find proof we're in Azure
        # in any version of the status json
        if any('DataSourceAzure' in x.get('datasource', '') for x in status_json.values()):
            # In azure, we rely on cloud services
            # instead of doing our own haproxy work
            nrpe_config.add_check(
                shortname='load-balancer',
                description='e2e check that regional load balancer is functioning',
                check_cmd=(
                    'check_http '
                    '--hostname ubuntu-archive-asm.trafficmanager.net '
                    '--url "/{}/" --string "dists/" '
                    '--expect \"200\\ OK\"'
                ).format(path_base),
            )

    nrpe_config.add_check(
        shortname='apache',
        description='Apache2 serving archive metadata',
        check_cmd=('check_http --hostname localhost --url "/{}/" ' '--string "dists/" --expect \"200\\ OK\"').format(
            path_base
        ),
    )
