#! /usr/bin/python3

import errno
import glob
import grp
import hashlib
import os
import pwd
import shutil
import socket
import traceback

from collections import OrderedDict
from datetime import datetime
from itertools import cycle

from subprocess import CalledProcessError, check_call, check_output, STDOUT

from .util import (
    FileSemaphore,
    REPO_SYNC_CRON_LOG,
    SemaphoreExistsError,
    SEMAPHORE_FILE,
)

import backoff


# The depth of the metadata history we keep around.
NUMBER_OF_REPOS_TO_KEEP = 2

# The number of times to retry rsync errors.
RSYNC_MAX_RETRIES = 4

# Number of times to retry when the upstream archive is in the middle of a
# sync.
UPSTREAM_SYNC_MAX_RETRIES = 4

# We try really hard to sync with the upstream hosts. Sometimes however we need
# to declare failure after trying to sync unsuccessfully with N hosts.
MAX_HOSTS_TO_TRY = 2

"""
!!! This file is kept clean of charmhelpers !!!

This file is run from a cronjob *outside* of the juju hook environment. Under
no circumstances should you import anything from charmhelpers in here.
"""


class ContentBad(Exception):
    '''Mirror content is inconsistent'''

    pass


class UpstreamMirrorSyncError(Exception):
    """The upstream mirror is currently syncing.

    We shouldn't sync with it right now since we will certainly get an invalid
    snapshot of its state."""


# VENDORING OF CHARMHELPERS FUNCTIONS.
# Some of the charmhelpers functions are needlessly reliant on being run in a
# hook environment. There are lifted and adapted: they are mostly thin wrappers
# around subprocess - ~tribaal


def rsync(from_path, to_path, options, log, flags="-r", check_output=check_output, user=None, timeout=None):
    """
    A wrapper around subprocess to execute an rsync between a source and a
    destination.

    Note: this was lifted from charm helpers to remove the useless dependency
    on hookenv via the not-injected log().

    @param check_output: An injection of subprocess.check_output
        (injectable for tests)

    This function has no side effects.
    """
    cmd = ["/usr/bin/rsync", flags]
    if timeout:
        cmd = ["/usr/bin/timeout", str(timeout)] + cmd
    if user:
        cmd = ["sudo", "-u", user] + cmd
    cmd.extend(options)
    cmd.append(from_path)
    cmd.append(to_path)
    log(" ".join(cmd))

    return check_output(cmd, stderr=STDOUT).decode('UTF-8').strip()


def symlink(source, destination, log=None, check_call=check_call):
    """
    A wrapper around subprocess to create a symlink between the source and the
    destination name.

    This function has no side effects.
    """
    if log:
        log("Symlinking {} as {}".format(source, destination))
    cmd = ["ln", "-sf", source, destination]
    check_call(cmd)


def mkdir(path, owner='root', group='root', perms=0o555, log=None):
    """Create a directory.

    The resulting directory is owned by the passed-in owner and group, and will
    have the passed-in permissions.

    This function has side-effects through os, pwd and grp
    """
    if log:
        log("Making dir {} {}:{} {:o}".format(path, owner, group, perms))
    uid = pwd.getpwnam(owner).pw_uid
    gid = grp.getgrnam(group).gr_gid
    realpath = os.path.abspath(path)
    if not os.path.exists(realpath):
        os.makedirs(realpath, perms)
    os.chown(realpath, uid, gid)
    os.chmod(realpath, perms)


def warning_log(log, message):
    """
    A simple wrapper around the passed-in log to make it output warning
    messages since all charmhelpers does is prepend WARNING: to them.

    This function has no side effects.
    """
    log("WARNING: {}".format(message))


# END OF VENDORING OF CHARMHELPERS FUNCTIONS.


def mirror_archive(hostname, apache_root, local_unit, log, mirror_series, rsync_module='ubuntu'):  # NOQA: C901
    """Sync Ubuntu repository metadata to the mirror.

    Metadata synchronization is performed and then the metadata is
    verified.

    Returns metadata version (destination file name)
    """

    now = datetime.utcnow().replace(microsecond=0)

    apache_data = '/'.join((apache_root, 'data'))
    link_active = '/'.join((apache_data, 'ubuntu_active'))
    link_next = '/'.join((apache_data, 'ubuntu_next'))
    dest = '/'.join((apache_data, 'ubuntu_{}_u{}'.format(now.isoformat(sep='_'), local_unit.split('/')[-1])))
    log('Performing sync to %s' % dest)

    # The archive may have multiple IP addresses, try each several times.
    archive_ips = [i for i in socket.gethostbyname_ex(hostname)[-1]]
    archive_ip_iterator = cycle(archive_ips)

    content_mirrored = False

    hosts_tried = 0

    while not content_mirrored and hosts_tried < MAX_HOSTS_TO_TRY:
        archive_ip = next(archive_ip_iterator)
        hosts_tried = hosts_tried + 1

        log('Archive readiness check for %s' % archive_ip)

        max_retries = 3
        sync_success = False
        for retry in range(1, max_retries + 1):
            try:
                # The following will be retried in case rsync fails, or in case the
                # upstream mirror is in the middle of a sync.
                _mirror_base(apache_root, archive_ip, dest, log=log, rsync_module=rsync_module)

                log('Sync from {} starting ({} of {}).'.format(archive_ip, retry, max_retries))
                # The following will be retried in case rsync fails.
                _mirror_metadata(archive_ip, dest, link_active, log, mirror_series, rsync_module=rsync_module)

                sync_success = True
                break
            except Exception:
                log(traceback.format_exc())
                log("Couldn't sync from host {} (attempt {} of {}).".format(archive_ip, retry, max_retries))

        if not sync_success:
            # If we get there it means the number of retries for this host was
            # reached.
            log("Couldn't sync from host {}; exhausted retries for current host.".format(archive_ip))
            continue

        log('Sync from {} completed.'.format(archive_ip))
        log('Starting repository mirror verification.')

        try:
            _verify_repository(dest, log)
        except ContentBad as excp:
            log('Verification failed: %s' % excp)
            log('Trying with next host.')
            continue

        log('Successful repository mirror verification.')
        content_mirrored = True

    if not content_mirrored:
        # We couldn't mirror from upstream for some reason. Let's cleanup what
        # we just downloaded, cleanup (why not).
        warning_log(log, 'Could not gather new metadata from archive.')
        clean_metadata_dir(apache_root, log)
        return None

    # Create a canary in the metadata directory to indicate a complete mirror
    open('_'.join((dest, 'good')), 'a').close()

    # link_active will not exist on first run, need to create it
    try:
        os.stat(link_active)
    except OSError as err:
        if err.errno is errno.ENOENT:
            symlink(dest, link_active)
        else:
            raise

    # The ubuntu_next link indicates the latest mirrored content
    try:
        os.stat(link_next)
    except OSError as err:
        if err.errno is errno.ENOENT:
            symlink(dest, link_next)
        else:
            raise

    # Delete all unused archive metadata mirror directories
    log('Removing old, unused repository mirrors.')
    clean_metadata_dir(
        apache_root,
        log,
        keep=set((os.path.realpath(dest), os.path.realpath('_'.join((dest, 'good'))), os.path.realpath(link_active))),
    )
    log('Unused repository mirrors removed.')

    return os.path.basename(dest)


def main(environment, log, mirror_archive=mirror_archive, check_call=check_call):
    """The main function for this script.

    This is called at the very bottom of this file should the file be executed
    as a script.
    """
    log("Performing metadata sync from cron.")
    sync_host = environment["SYNC_HOST"]
    try:
        rsync_module = environment["RSYNC_MODULE"]
    except KeyError:
        rsync_module = 'ubuntu'
    apache_root = environment["APACHE_ROOT"]
    local_unit = environment["LOCAL_UNIT"]
    unit_path = environment["UNIT_PATH"]
    mirror_series = environment["MIRROR_SERIES"]

    meta_ver = mirror_archive(sync_host, apache_root, local_unit, log, mirror_series, rsync_module=rsync_module)
    if meta_ver is None:
        # If we failed to download and/or validate a new version of the
        # archives, don't propagate the information to other units.
        # This can happen when we give up retrying the validation of the latest
        # version - for example when the upstream archive is syncing itself.
        log("Won't notify peers because latest version is None. Bailing out.")
        return None

    # Call the sync hook inside the juju environment to propagate changes to
    # other units
    charm_env = '/var/lib/juju/agents/{}/charm/bin/charm-env'.format(unit_path)
    path = '{} /var/lib/juju/agents/{}/charm/hooks/ubuntu-repository-cache-sync {}'
    path = path.format(charm_env, unit_path, meta_ver)
    cmd = ['juju-run', local_unit, path]

    log("Executing hook: {}".format(cmd))

    check_call(cmd)

    return meta_ver


def rsync_should_giveup(exception):
    """Based on the provided returncode, should rsync be retried or aborted.

    This function has no side-effects.
    """

    returncode = exception.returncode
    # Exit values for rsync that will trigger a retry of the operation
    exitval_retry = frozenset(
        (
            5,  # Error starting client-server protocol
            6,  # Daemon unable to append to log-file
            10,  # Error in socket I/O
            11,  # Error in file I/O
            12,  # Error in rsync protocol data stream
            13,  # Errors with program diagnostics
            14,  # Error in IPC code
            20,  # Received SIGUSR1 or SIGINT
            21,  # Some error returned by waitpid()
            22,  # Error allocating core memory buffers
            23,  # Partial transfer due to error
            24,  # Partial transfer due to vanished source files
            25,  # The --max-delete limit stopped deletions
            30,  # Timeout in data send/receive
            35,  # Timeout waiting for daemon connection
        )
    )

    if returncode in exitval_retry:
        return False  # Don't give up

    return True  # Do give up by default.


def clean_metadata_dir(apache_root, log, keep=None):
    '''Delete non-active metadata directories'''

    log('Cleaning metadata directory')
    apache_data = '/'.join((apache_root, 'data'))
    if not keep:
        active = os.path.realpath('/'.join((apache_data, 'ubuntu_active')))
        canary = '_'.join((active, 'good'))
        keep = set((active, canary))
        log('ubuntu_active currently points at: {}'.format(active))
    else:
        log('Keeping list provided: {}'.format(', '.join(keep)))

    glob_path = '/'.join((apache_data, '/ubuntu_[0-9]*'))

    # Let's keep the last two good repository metadata caches. This saves us
    # lots of time when we need to revert, instead of 15-20 mins to wait for a
    # complete sync.
    count = 0
    entries = sorted([os.path.realpath(p) for p in glob.glob(glob_path)], reverse=True)
    for entry in entries:
        if count < NUMBER_OF_REPOS_TO_KEEP and entry.endswith('_good'):
            keep.add(entry[:-5])
            keep.add(entry)
            count += 1

    for entry in [os.path.realpath(p) for p in glob.glob(glob_path)]:
        if entry not in keep:
            # We want to make sure it isn't the directory
            # ubuntu_active points to (LP#1799595).
            active = os.path.realpath('/'.join((apache_data, 'ubuntu_active')))
            if entry == active:
                log('ubuntu_active points to {}, skipping'.format(entry))
            elif os.path.isfile(entry):
                # XXX: Should be a debug
                log('Removing file {}'.format(entry))
                os.remove(entry)
            elif os.path.isdir(entry) and not os.path.islink(entry):
                # XXX: Should be a debug
                log('Removing tree {}'.format(entry))
                shutil.rmtree(entry)
        else:
            log("Keeping {}".format(entry))


@backoff.on_exception(backoff.expo, CalledProcessError, max_tries=RSYNC_MAX_RETRIES, giveup=rsync_should_giveup)
def _mirror_metadata(archive_ip, dest, link_dest, log, mirror_series, rsync_module='ubuntu'):
    '''Mirror metadata from the archive.'''

    # Perform an rsync of the metadata, excluding some contents
    rsync_source = '{}::{}/'.format(archive_ip, rsync_module)
    rsync_options = (
        '--quiet',
        '--recursive',
        '--links',
        '--perms',
        '--chmod=g+w',
        '--times',
        '--compress',
        '--delete',
        '--delete-during',
        '--link-dest=%s/' % link_dest,
        '--exclude=.trace',
        '--exclude=indices',
        '--exclude=project/trace',
        '--exclude=pool',
        '--exclude=dists/*/*/debian-installer',
        '--exclude=dists/*/*/dist-upgrader-all',
        '--exclude=dists/*/*/installer*',
        '--exclude=dists/*/*/uefi',
    )
    # Ordering of the include/exclude rules is important, rsync will
    # follow the first matching rule. So place the common rules above
    # the rules to limit by series.
    if mirror_series:
        for series in mirror_series.split():
            rsync_options += ('--include=dists/%s*' % series,)
        # In order to limit to just the requests series, we must explicitly
        # exclude everything else under dists as the last rule.
        rsync_options += ('--exclude=dists/*',)

    rsync(rsync_source, dest, rsync_options, log, user='www-sync', timeout=3600)


@backoff.on_exception(backoff.expo, CalledProcessError, max_tries=RSYNC_MAX_RETRIES, giveup=rsync_should_giveup)
@backoff.on_exception(backoff.expo, UpstreamMirrorSyncError, max_tries=UPSTREAM_SYNC_MAX_RETRIES, factor=60)
def _mirror_base(apache_root, archive_ip, dest, log=None, rsync_module='ubuntu'):
    """Mirror just the root of the archive.

    This checks for the file 'Archive-Update-in-Progress-*' and
    throws ContentBad if the archive source isn't ready.
    Use this as a quick test prior to attempting a full mirror
    operation.
    """

    source = '{}::{}/'.format(archive_ip, rsync_module)
    mkdir(dest, owner='www-sync', group='www-data', perms=0o775, log=log)
    apache_data = '/'.join((apache_root, 'data'))
    link_dest = '/'.join((apache_data, 'ubuntu_active'))

    # Sync just the root directory
    rsync_options = (
        '--links',
        '--perms',
        '--times',
        '--dirs',
        '--chmod=g+w',
        '--delete',
        '--compress',
        '--link-dest=%s/' % link_dest,
    )
    rsync(source, dest, rsync_options, log, flags='--quiet', user='www-sync', timeout=3600)

    # Check for file 'Archive-Update-in-Progress*' which signals
    # the need to sleep or try another IP
    if glob.glob('/'.join((dest, 'Archive-Update-in-Progress-*'))):
        raise UpstreamMirrorSyncError('Remote archive update in progress.')


def _verify_repository(archive_root, log):
    '''Verify the content of a Ubuntu software repository's metadata'''

    dists_root = '/'.join((archive_root, 'dists'))

    # Traverse distributions and fail early by starting with the newest.
    # The newest (development branch) will have the highest rate of churn,
    # therefore the greatest chance of being invalid.
    dists = sorted(
        [
            ent
            for ent in os.listdir(dists_root)
            if os.path.isdir('/'.join((dists_root, ent))) and not os.path.islink('/'.join((dists_root, ent)))
        ],
        reverse=True,
    )

    for dist in dists:
        log('Verifying distribution %s' % dist)
        dist_dir = '/'.join((dists_root, dist))
        release = _parse_release_file(dist_dir)
        _verify_distribution(dist_dir, release)

    for uip in glob.glob('/'.join((archive_root, 'Archive-Update-in-Progress-*'))):
        # XXX: Should be debug-only
        log('Removing {} from clean repository'.format(uip))
        os.remove(uip)


def _parse_release_file(path):
    """Read a distribution Release file located at 'path'

    Returns a dictionary describing Release file contents.
    """

    release = {}
    release['name'] = os.path.basename(path.strip('/'))
    release['mtime'] = os.stat(path + '/Release').st_mtime

    # Collect expected SHA256 sum values for files in the distribution
    files = release['files'] = OrderedDict({})
    with open(path + '/Release', 'r') as release_file:
        found = False
        for line in release_file:
            if not found:
                if line.startswith('SHA256:'):
                    found = True
            else:
                if line.startswith(' '):
                    line = line.strip().split()
                    files[line[2]] = {'SHA256': line[0], 'size': line[1]}
                else:
                    break
    return release


def _verify_distribution(path, release):
    '''Check distribution file sizes/SHA256sums against Release file values'''

    files = release['files']
    for filename in files:
        filepath = '/'.join((path, filename))

        try:
            disk_size = os.stat(filepath).st_size
        except OSError as excp:
            if excp.errno is errno.ENOENT:
                # There will be a number of files listed in the Release
                # file that are not present on disk.  Alternatively, we
                # could walk release tree and verify against the release
                # metadata
                continue
            else:
                raise

        size = int(files[filename]['size'])
        if disk_size != size:
            msg = 'Sizes do not match for {} in {}'.format(filename, release['name'])
            raise ContentBad(msg)

        with open(filepath, 'rb') as infile:
            sha256 = hashlib.sha256()
            while True:
                buf = infile.read(1024 * 1024 * 100)
                if not buf:
                    break
                sha256.update(buf)
        digest = sha256.hexdigest()
        if digest != files[filename]['SHA256']:
            file_mtime = os.stat(filepath).st_mtime
            msg = 'SHA256 mismatch for {} in {}.'.format(filename, release['name'])
            if release['mtime'] < file_mtime:
                ' '.join((msg, 'Stale Release file.'))
            raise ContentBad(msg)


if __name__ == "__main__":

    # Use the logging infrastructure to print messages with timestamps to
    # stdout. This is a cheap way to get good timestamps. Ideally we should
    # move to leverage the logging infrastructure better. ~tribaal
    import logging

    file_handler = logging.FileHandler(REPO_SYNC_CRON_LOG)
    file_handler.setLevel(logging.INFO)

    basic_formatter = logging.Formatter('%(asctime)s - %(message)s')
    file_handler.setFormatter(basic_formatter)

    cache_logger = logging.getLogger('ubuntu-repository-cache')
    cache_logger.setLevel(logging.INFO)
    cache_logger.addHandler(file_handler)

    backoff_logger = logging.getLogger('backoff')
    backoff_logger.setLevel(logging.INFO)
    backoff_logger.addHandler(file_handler)

    environment = os.environ

    try:
        with FileSemaphore(SEMAPHORE_FILE):
            cache_logger.info("Metadata sync started from metadata_sync.py:__main__")
            main(environment, cache_logger.info)
            cache_logger.info("Metadata sync from __main__ ended.")
    except SemaphoreExistsError:
        cache_logger.info("Can't acquire semaphore: {} already exists.".format(SEMAPHORE_FILE))
