'''Functions relating to mirrored Ubuntu package repository content'''

import os
import sys
import time
import traceback

from subprocess import (
    CalledProcessError,
)

from charmhelpers.contrib import unison
from charmhelpers.contrib.hahelpers import cluster

from charmhelpers.core import (
    hookenv,
    host,
    templating,
    unitdata,
)

from ubuntu_repository_cache import service
from ubuntu_repository_cache import util
from ubuntu_repository_cache import metadata_sync

LOG = hookenv.log
SERVICE = 'ubuntu-repository-cache'


def _rsync(source, dest, link_dest=None, optlist=None, user='www-sync', retries=1):
    '''Perform an rsync as a specified user'''

    options = [
        '--quiet',
        '--recursive',
        '--links',
        '--perms',
        '--chmod=g+w',
        '--times',
        '--compress',
        '--delete',
        '--delete-before',
        '--timeout=120',
        '-e',
        'sudo -u {} ssh -l {}'.format(user, user),
    ]
    if optlist:
        options.extend(optlist)
    if link_dest:
        options.append('--link-dest=%s/' % link_dest)
    attempt = 0
    while True:
        try:
            host.rsync(source, dest, options=options, timeout=3600)
        except CalledProcessError as excp:
            LOG('rsync failed({}): {}'.format(excp.returncode, excp.output), hookenv.WARNING)
            if not metadata_sync.rsync_should_giveup(excp):
                attempt += 1
                if attempt < retries:
                    LOG('Retrying rsync. (attempt #%s)' % attempt)
                    time.sleep(10)
                    continue
                else:
                    LOG('No retries left.')
                    raise
            else:
                LOG('rsync failed({}): {}'.format(excp.returncode, excp.output), hookenv.WARNING)
                raise
        return


@util.run_as_user('www-sync')
def rsync_to_peers(source_dir, dest_dir, link_dir=None, user='www-sync', peers=None):
    """
    Use rsync to push a directory to all configured peers.

    Peers are configured in the 'cluster' relationship using
    charmhelpers.contrib.unison.ssh_authorized_peers(); this list can be
    overridden with the peers parameter.

    A list of the peers whose transfers were successful will be
    returned.
    """

    if peers is None:
        peers = unison.collect_authed_hosts(peer_interface='cluster')

    successful_peers = []
    for peer in peers:
        rsync_dest = '{}@{}:{}'.format(user, peer, dest_dir)
        LOG('Syncing {} to {}.'.format(source_dir, rsync_dest))
        try:
            _rsync(source_dir, rsync_dest, link_dest=link_dir, user=user, retries=10)
        except Exception:
            LOG('Sync to {} failed'.format(peer))
            traceback.print_exc(file=sys.stdout)
        else:
            successful_peers.append(peer)
            LOG('Sync to {} complete.'.format(peer))
    return successful_peers


@util.run_as_user('www-sync')
def rsync_from_peer(source_dir, dest_dir, peer_ip, link_dir=None, user='www-sync'):
    '''Use rsync to pull a directory from a configured peer.  '''

    rsync_source = '{}@{}:{}'.format(user, peer_ip, source_dir)
    LOG('Syncing {} to {}.'.format(rsync_source, dest_dir))
    _rsync(rsync_source, dest_dir, link_dest=link_dir, user=user, retries=10)
    LOG('Sync to {} complete.'.format(peer_ip))


def _leader_update_metadata():
    """Update the metadata on a leader in the cluster.

    If the leader is newly elected, it will pull down the latest
    metadata.  The metadata is not pushed to peers.  The peers will
    pull the metadata once on the next relation-changed hook triggered
    by setting the meta_version relation data.
    """

    LOG('Updating metadata on the leader')
    # Relation data
    rel = hookenv.relations()[hookenv.relation_type()][hookenv.relation_id()]

    # Read the leader ID from Juju. This is extremely likely to be "this unit"
    # but there is a tiny race window if the leader changes between
    # peer_update_metadata's check and now. ~tribaal
    leader_id = hookenv.leader_get(attribute="leader_id")
    # Log the race if it happened:
    if leader_id != hookenv.local_unit():
        LOG('Leader changed between peer_update_metadata and ' '_leader_update_metadata', hookenv.WARNING)

    leader_rel = rel[leader_id]

    active_ver = util.get_active_metaver()
    meta_ver = leader_rel.get('meta_version')
    apache_root = unitdata.kv().get('apache-root')
    apache_data = '/'.join((apache_root, 'data'))
    if meta_ver:
        meta_dir = '/'.join((apache_data, meta_ver))
        canary = '_'.join((meta_dir, 'good'))

    # Common case.  All good, just need to update to latest metadata
    if meta_ver and os.path.exists(meta_dir) and os.path.exists(canary):
        if meta_ver != active_ver:
            LOG('Updated to metadata version {}'.format(meta_ver))
            util.update_active_symlink(meta_ver)
        else:
            LOG('Nothing to update')
    else:
        LOG('Leader missing expected metadata. Recovering.')
        LOG('meta_ver:{} active_ver:{}'.format(meta_ver, active_ver), hookenv.DEBUG)
        metadata_sync.clean_metadata_dir(apache_root, LOG)

        # Don't use the existing metadata.  If this is a current unit
        # taking over from a prior leader, use of the existing metadata
        # could move units back in time.
        sync_host = hookenv.config()['sync-host']
        rsync_module = hookenv.config()['rsync-module']
        mirror_series = hookenv.config()['mirror-series'].strip()

        # This context manager will create a lock file to prevent cronjobs
        # from running while the initial upstream sync is running. In the
        # unlikely even that the lock is already in the filesystem, it is
        # reasonable to panic and just explode (hence the absence of try block)
        with util.FileSemaphore(util.SEMAPHORE_FILE):
            new_ver = metadata_sync.mirror_archive(
                sync_host, apache_root, hookenv.local_unit(), LOG, mirror_series, rsync_module=rsync_module
            )

        if not new_ver:
            LOG('Leader could not gather new metadata', hookenv.WARNING)
            return
        util.update_active_symlink(new_ver)

        LOG('Updating metadata version to {}'.format(new_ver), hookenv.DEBUG)
        hookenv.relation_set(meta_version=new_ver)

    service.unpause()
    metadata_sync.clean_metadata_dir(apache_root, LOG)
    LOG('Metadata update complete.')


def _nonleader_update_metadata():
    '''Update the metadata on a non-leader peer in the cluster'''

    LOG('Updating metadata on a peer')

    rel = hookenv.relations()[hookenv.relation_type()][hookenv.relation_id()]
    local_rel = rel[hookenv.local_unit()]

    # Read the leader ID from Juju. This is extremely likely to be another unit
    # than the current one, but there is a tiny race window if the leader
    # changes between peer_update_metadata's check and now. ~tribaal
    leader_id = hookenv.leader_get(attribute="leader_id")
    # Log the race if it happened:
    if leader_id == hookenv.local_unit():
        LOG('Leader changed between peer_update_metadata and ' '_nonleader_update_metadata', hookenv.WARNING)

    if leader_id not in rel:
        # Peers can join in any order and the leader may not have joined yet.
        # Log who we think is the leader and return until it joins
        LOG('Leader {} not yet related, can\'t complete ' '_nonleader_update_metadata'.format(leader_id))
        return

    leader_rel = rel[leader_id]

    active_ver = util.get_active_metaver()
    meta_ver = leader_rel.get('meta_version')
    apache_root = unitdata.kv().get('apache-root')
    apache_data = '/'.join((apache_root, 'data'))
    if meta_ver:
        meta_dir = '/'.join((apache_data, meta_ver))
        canary = '_'.join((meta_dir, 'good'))

    # Clean up meta_version on local relation if no longer the leader
    if local_rel.get('meta_version'):
        LOG('Clearing metadata', hookenv.DEBUG)
        hookenv.relation_set(meta_version=None)

    if not leader_rel.get('ssh_pub_key'):
        LOG('SSH keys not yet present for leader.', hookenv.INFO)
        return
    if not meta_ver:
        LOG('Metadata version not yet provided by leader.', hookenv.INFO)
        return

    # Need to update metadata
    if meta_ver and meta_ver != active_ver:
        # Already have good metadata
        if os.path.exists(meta_dir) and os.path.exists(canary):
            LOG('Updated to metadata version {}'.format(meta_ver))
            util.update_active_symlink(meta_ver)
            service.unpause()

        # Need to pull metadata from leader
        else:
            # Pause service and make space for new metadata
            service.pause()
            metadata_sync.clean_metadata_dir(apache_root, LOG)

            # Can we connect to the leader?
            leader_hosts = leader_rel.get('ssh_authorized_hosts')
            if not leader_hosts or hookenv.unit_private_ip() not in leader_hosts:
                LOG('Leader has not added IP to authorized_hosts yet, ' 'can not update.')
                return

            # Pull metadata from the leader if it exists and is ready
            leader_ip = leader_rel.get('private-address')
            if util.remote_path_exists('www-sync', leader_ip, canary):
                LOG('Pulling metadata {} from leader.'.format(meta_ver))
                active_dir = '/'.join((apache_data, 'ubuntu_active'))
                source = meta_dir + '/'
                dest = meta_dir
                try:
                    rsync_from_peer(source, dest, leader_rel.get('private-address'), link_dir=active_dir)
                except CalledProcessError:
                    traceback.print_exc()
                    LOG('Failed to pull metadata from leader; service stopped' ' until next update from leader.')
                else:
                    util.update_active_symlink(meta_ver)
                    service.unpause()

            # Do nothing. Service is stopped until next update from leader
            else:
                LOG('Metadata dir does not exist on leader.', hookenv.DEBUG)
    metadata_sync.clean_metadata_dir(apache_root, LOG)
    LOG('Metadata update complete.')


def peer_update_metadata():
    """Update the metadata on a clustered peer to the current version.

    If the 'cluster' relationship exists, this will look for the
    current 'meta_version' key on the leader and attempt to
    update the active metadata to this version.

    This function should be called from the cluster-relation-joined,
    cluster-relation-changed, and cluster-relation-departed hooks.
    """

    if (
        not hookenv.is_relation_made('cluster')
        or not hookenv.in_relation_hook()
        or not hookenv.relation_type() == 'cluster'
    ):
        return

    if cluster.is_elected_leader(None):
        # Leader does not pause, a newly appointed leader must be available
        if util.in_failover():
            service.unpause()
        _leader_update_metadata()
    else:
        _nonleader_update_metadata()


def update_checks(nrpe_config):
    '''Update nagios check for mirror metadata age'''

    service_script = '/'.join((hookenv.charm_dir(), 'files', 'nrpe-external-master', 'check_metadata_age'))
    metadata_dir = '/'.join((unitdata.kv().get('apache-root'), 'data', 'ubuntu_active'))
    check_metadata_age_config_path = '/srv/{}/var/check_metadata_age'.format(hookenv.service_name())

    config = hookenv.config()
    context = {}
    context['crit_age'] = config['sync-age-crit']
    context['warn_age'] = config['sync-age-warn']
    templating.render('check_metadata_age-config', check_metadata_age_config_path, context)

    nrpe_config.add_check(
        shortname='mirror',
        description='Ubuntu repository cache metadata age',
        check_cmd=' '.join((service_script, metadata_dir, check_metadata_age_config_path)),
    )
