'''Miscellaneous helpers for the ubuntu-repository-cache charm '''

import glob
import os
import random
import subprocess
import time

from collections import OrderedDict

from charmhelpers import fetch
from charmhelpers.contrib.charmsupport import nrpe
from charmhelpers.core import (
    hookenv,
    host,
    templating,
    unitdata,
)
from charmhelpers.contrib.hahelpers import cluster

from ubuntu_repository_cache import apache
from ubuntu_repository_cache import squid
from ubuntu_repository_cache import storage
from ubuntu_repository_cache import util


CONFIG_DIR = '/etc/repository-sync'
CONFIG_PATH = os.path.join(CONFIG_DIR, 'ubuntu-repository-cache')
LOG = hookenv.log
SERVICE = 'ubuntu-repository-cache'


@util.run_as_user('root')
def write_status(state):
    '''Create a status file in /srv/<service_name>/var/ with simple state'''

    var_dir = '/'.join(('/srv', hookenv.service_name(), 'var'))
    state_filename = '/'.join((var_dir, hookenv.service_name()))

    if not os.path.exists(var_dir):
        host.mkdir(var_dir)
    with open(state_filename, 'w') as state_file:
        content = 'STATE={}\n'.format(state)
        state_file.write(content)


def close_ports():
    '''Close all ports for the service'''
    hookenv.close_port(80)


def open_ports():
    '''Open all ports for the service'''
    hookenv.open_port(80)


def install():
    '''Perform service installation'''

    config = hookenv.config()

    # We can allow for pre-install scripts to alter the operating
    # environment, it doesn't mean we can support it.
    preinstall_files = glob.glob('exec.d/*/charm-pre-install')
    if preinstall_files:
        LOG('============================================================', hookenv.CRITICAL)
        LOG('Executing pre-install script(s).', hookenv.CRITICAL)
    for pre in preinstall_files:
        if not os.path.isfile(pre) or not os.access(pre, os.X_OK):
            LOG('Can not execute %s' % pre, hookenv.CRITICAL)
            raise RuntimeError('Pre-install script %s not executable' % pre)
        try:
            LOG('============================================================', hookenv.CRITICAL)
            LOG('Calling %s' % pre, hookenv.CRITICAL)
            cmd = pre
            subprocess.check_call(cmd)
        except subprocess.CalledProcessError as xcp:
            LOG('%s failed, retcode %d' % (pre, xcp.returncode), hookenv.CRITICAL)
            raise RuntimeError('%s failed, retcode %d' % (pre, xcp.returncode))
        else:
            LOG('%s succeeded' % pre, hookenv.CRITICAL)
            LOG('============================================================', hookenv.CRITICAL)

    LOG('Installing services', hookenv.INFO)
    write_status('installing')
    # Change apt/sources.list to use the main archive URL.
    # This charm may be provide the ubuntu package repository for
    # the cloud where it is being deployed.  In that case the
    # URL for the repository in the cloud image will point to the
    # service provided by this charm and it could be unavailable.
    if config['update-unit-apt-sources']:
        url = 'http://{}/{}/'.format(config['sync-host'], config['path-base'])
        LOG('Changing apt source list url to %s' % url)
        # XXX: to fix as python3-apt isn't a thing.
        util.change_aptsources_url(url)

    fetch.apt_update(fatal=True)

    storage.setup()
    apache.install()
    squid.install()
    try:
        subprocess.check_call(
            ['useradd', '--gid', 'www-data', '--create-home', '--no-user-group', '--shell', '/bin/bash', 'www-sync']
        )
    except subprocess.CalledProcessError as e:
        if e.returncode == 9:  # User already exists.
            pass
        else:
            raise

    # Make sure a log file is available for the user to write to - this will be
    # use by the cron, run as that user.
    subprocess.check_call(['install', '-o', 'www-sync', '-d', util.REPO_SYNC_DIR])
    subprocess.check_call(['touch', util.REPO_SYNC_CRON_LOG])
    subprocess.check_call(['chown', 'www-sync', util.REPO_SYNC_CRON_LOG])

    write_status('installed')


def pause():
    """Temporarily route metadata requests to a peer (non-leader)

    Use this when the metadata does not exist so that users don't
    find an empty mirror.  This is achieved by using the next peer
    as a proxy for metadata.  In the case that this unit is the leader,
    no proxy is used.

    Don't stop Squid, because it:
    * takes a long time to stop/start,
    * trashes the in-memory cache when stopped, and
    * can continue serving requests if left running.
    """

    # Don't pause on leader, but warn developer of incorrect usage
    # Change servicing of metadata by forwarding requests for
    # metadata
    if not cluster.is_elected_leader(None):
        LOG('Pausing services', hookenv.INFO)
        write_status('paused')
        util.set_failover()
        render_configs()
    else:
        LOG('Pause should not be called on the leader, report to support', hookenv.DEBUG)


def unpause():
    '''Make the service available after being paused by pause_service()'''

    LOG('Un-pausing services', hookenv.INFO)
    if not host.service_running('squid-deb-proxy'):
        squid.start()

    # Change servicing of metadata to use local metadata
    util.unset_failover()
    render_configs()
    apache.start()
    open_ports()
    write_status('running')


def stop():
    '''Stop service and prevent from running in the future'''

    LOG('Stopping services', hookenv.INFO)
    close_ports()
    apache.stop()
    squid.stop()
    write_status('stopped')


def start():
    '''Start the service'''

    LOG('Starting services', hookenv.INFO)
    squid.start()
    apache.start()
    open_ports()
    write_status('running')


def is_running():
    '''Check if service is running'''

    return apache.service_running() and squid.service_running()


def service_map():
    """Mapping of configuration files to services.

    This is used by charmhelpers.core.host.restart_on_change to
    restart services when configuration files are altered.
    """

    return {
        '/etc/squid-deb-proxy/squid-deb-proxy.conf': ['squid-deb-proxy'],
        '/etc/squid-deb-proxy/allowed-networks-src.acl': ['squid-deb-proxy'],
        ('/etc/squid-deb-proxy/mirror-dstdomain.acl.d/' '99-ubuntu-repository-cache'): ['squid-deb-proxy'],
        '/etc/apache2/sites-available/archive_ubuntu_com.conf': ['apache2'],
    }


def reload_on_change(mapping, restart_on_failure=False):
    """Reload services based on configuration files changing

    (Based on charmhelpers.core.host.restart_on_change)
    This function is used as a decorator, for example::
        @reload_on_change({
            '/etc/ceph/ceph.conf': [ 'cinder-api', 'cinder-volume' ]
            })
        def ceph_client_changed():
            pass  # your code here

    In this example, the cinder-api and cinder-volume services
    would be reloaded if /etc/ceph/ceph.conf is changed by the
    ceph_client_changed function.
    """

    def wrap(function):
        def wrapped_f(*args):
            checksums = {}
            for path in mapping:
                checksums[path] = host.file_hash(path)
            function(*args)
            restarts = []
            for path in mapping:
                if checksums[path] != host.file_hash(path):
                    restarts += mapping[path]
            services_list = list(OrderedDict.fromkeys(restarts))
            for service_name in services_list:
                LOG('Reloading %s service' % service_name, hookenv.DEBUG)
                host.service_reload(service_name, restart_on_failure=restart_on_failure)

        return wrapped_f

    return wrap


@reload_on_change(service_map())
def render_configs():
    '''Render the configuration templates for all required services'''

    LOG('Rendering configuration templates')
    config = hookenv.config()

    apache.render_configs()
    squid.render_configs()

    # Log file rotation
    logrotate_filename = '/etc/logrotate.d/apache2'
    logrotate_context = {
        'logrotate_rotate': config['logrotate_rotate'],
        'logrotate_count': config['logrotate_count'],
        'logrotate_dateext': config['logrotate_dateext'],
    }
    templating.render('logrotate/logrotate.conf.template', logrotate_filename, logrotate_context, perms=0o644)

    if not os.path.exists(CONFIG_DIR):
        host.mkdir(CONFIG_DIR)
    urc_config_path = CONFIG_PATH

    # Cron job for metadata update on leader
    cron_filename = '/etc/cron.d/ubuntu-repository-cache_rsync'
    if cluster.is_elected_leader(None):
        cron_context = {}
        if not unitdata.kv().get('rsync-minutes'):
            # Don't schedule an rsync within 15 minutes of startup
            minutes = (random.randint(0, 43) + time.gmtime().tm_min + 15) % 60
            unitdata.kv().set('rsync-minutes', minutes)
        cron_context['ApacheRoot'] = unitdata.kv().get('apache-root')
        cron_context['SyncHost'] = config['sync-host']
        cron_context['PathBase'] = config['path-base']
        cron_context['RsyncModule'] = config['rsync-module']
        cron_context['Minutes'] = unitdata.kv().get('rsync-minutes')
        cron_context['UnitId'] = hookenv.local_unit()
        cron_context['UnitPath'] = '-'.join(('unit', hookenv.local_unit().replace('/', '-')))
        cron_context['MirrorSeries'] = config['mirror-series'].strip()
        cron_context['ConfigPath'] = urc_config_path
        cron_context['CharmEnvPath'] = os.path.join(hookenv.charm_dir(), 'bin', 'charm-env')
        templating.render('cron/ubuntu-repository-cache_rsync.cron', cron_filename, cron_context)
        templating.render('cron/ubuntu-repository-cache-config', urc_config_path, cron_context)
    else:
        try:
            os.stat(cron_filename)
        except OSError:
            pass
        else:
            os.remove(cron_filename)


def update_checks():
    '''Update Nagios checks for all services and add overall metadata check'''
    # Dynamic import - prevents circular import problems.
    from ubuntu_repository_cache import mirror

    for relid in hookenv.relation_ids('nrpe-external-master'):
        hookenv.relation_set(relation_id=relid, primary=True)
    nrpe_config = nrpe.NRPE()

    # Services
    apache.update_checks(nrpe_config)
    squid.update_checks(nrpe_config)
    storage.update_checks(nrpe_config)
    mirror.update_checks(nrpe_config)

    # Metadata service check
    service_script = '/'.join((hookenv.charm_dir(), 'files', 'nrpe-external-master', 'check_metadata_service'))
    var_dir = '/'.join(('/srv', hookenv.service_name(), 'var'))
    state_filename = '/'.join((var_dir, hookenv.service_name()))

    nrpe_config.add_check(
        shortname='metadata',
        description='Ubuntu repository cache metadata service',
        check_cmd=' '.join((service_script, state_filename)),
    )

    nrpe_config.write()
