'''Functions related to squid-deb-proxy use within this charm'''

import os
import subprocess
import re

from charmhelpers import fetch

from charmhelpers.core import (
    hookenv,
    host,
    templating,
    unitdata,
)

from charmhelpers.core.host import lsb_release

from . import util

LOG = hookenv.log
SERVICE = 'squid-deb-proxy'


@util.run_as_user('proxy')
def init_caches():
    """Initialize the squid on-disk caches

    This must be called while the service is stopped and after the
    configuration has been written.
    """

    if not unitdata.kv().get('squid-disk-caches'):
        LOG('No on disk cache directories founds for squid.', hookenv.ERROR)
        return
    subprocess.check_call(['/usr/sbin/squid3', '-N', '-z', '-f', '/etc/squid-deb-proxy/squid-deb-proxy.conf'])


def size_caches(system_total_memory=0):
    """Determine the sizes for in-memory/on-disk squid caches."""

    config = hookenv.config()
    avail_disk = config.get('cache-storage-size', 0)
    if not avail_disk:
        for cache in unitdata.kv().get('squid-disk-caches'):
            avail_disk += cache[1]
        LOG('{}MB of disk cache available for squid.'.format(avail_disk), hookenv.DEBUG)

    memory = system_total_memory
    if not system_total_memory:
        memory = util.system_total_mem()
    # Set aside 30% of system memory for squid memory cache.
    memory = memory * 0.30
    # Estimate of 100MB of overhead for squid
    memory = max(memory - 100, 0)

    if memory < 128:
        LOG('Squid has too little memory ({}MB), ' 'can not deploy service'.format(memory), hookenv.ERROR)
        raise RuntimeError('Too little system memory for squid deployment')

    # Start by giving squid a maximum of 256MB for in-memory cache
    squid_cache_mem = min(memory, 256)
    memory -= squid_cache_mem

    # Estimate 20MB of memory overhead per GB of on-disk cache
    max_disk = int(memory / 20.0 * 1024)
    squid_cache_disk = min(max_disk, avail_disk)
    disk_overhead = int(round(squid_cache_disk / 1024.0 * 20))
    memory -= disk_overhead

    # Remaining memory (up to 1/2 total RAM) is given to in-memory cache
    squid_cache_mem += memory
    squid_cache_mem = int(squid_cache_mem)

    unitdata.kv().set('squid-cache-mem', squid_cache_mem)
    unitdata.kv().set('squid-cache-disk', squid_cache_disk)
    LOG(
        'Squid sizing complete. In-memory cache: {}MB '
        'On-disk cache: {}MB'.format(unitdata.kv().get('squid-cache-mem'), unitdata.kv().get('squid-cache-disk')),
        hookenv.INFO,
    )


def install():
    '''Perform the install/config and disable avahi'''

    LOG('Installing squid', hookenv.INFO)
    fetch.apt_install(SERVICE, fatal=True)

    release = lsb_release()['DISTRIB_RELEASE']

    # We don't need avahi-daemon ...
    try:
        host.service_pause('avahi-daemon')
    except ValueError:
        # ValueError: Unable to detect avahi-daemon as SystemD, Upstart
        # /etc/init/avahi-daemon.conf or SysV /etc/init.d/avahi-daemon
        # If avahi-daemon isn't installed, do nothing
        pass

    # ... or squid3 running
    if release < '16.04':  # before xenial
        squid_service = 'squid3'
    else:
        squid_service = 'squid'

    host.service_pause(squid_service)
    stop()

    render_configs()
    init_caches()


@util.run_as_user('root')
def start():
    '''Start the service and remove the override file blocking startup'''

    LOG('Starting squid', hookenv.INFO)
    override_path = '/etc/init/{}.override'.format(SERVICE)
    if os.path.exists(override_path):
        os.remove(override_path)

    host.service_start(SERVICE)


@util.run_as_user('root')
def stop():
    '''Stop the service and block startup at boot'''

    LOG('Stopping squid', hookenv.INFO)
    with open('/etc/init/{}.override'.format(SERVICE), 'w+') as override_file:
        override_file.write('manual\n')
    host.service_stop(SERVICE)


@util.run_as_user('root')
def service_running():
    '''Return True is the service is running'''

    return host.service_running(SERVICE)


def peer_config():
    '''Return the configuration strings for squid cache peers'''

    config = ''
    for ip in util.peer_IPs():
        config = '\n'.join((config, 'cache_peer %s sibling 8000 3130' % ip))
    return config


def cache_dirs():
    """Generate 'cache_dir' lines for the squid configuration file.

    Returns a string containing the cache_dir line(s)
    """
    cache_str = ''
    config = hookenv.config()
    max_size = unitdata.kv().get('squid-cache-disk')
    for path, free in unitdata.kv().get('squid-disk-caches'):
        if config.get('cache-storage-size', 0):
            size = int(config['cache-storage-size'])
        else:
            size = int(min(max_size, free))

        max_size -= size
        cache_str = ''.join((cache_str, 'cache_dir aufs {} {} 16 256\n'.format(path, size)))
    return cache_str


def cache_mem(system_total_memory=0):
    """Generate 'cache_mem' lines for the squid configuration file.

    Returns a string containing the cache_mem config
    """
    size_caches(system_total_memory)

    config = hookenv.config()
    cache_mem = config.get('cache-memory-size', 0)
    if not cache_mem:
        cache_mem = unitdata.kv().get('squid-cache-mem')

    cache_str = 'cache_mem {} MB'.format(str(cache_mem))
    return cache_str


@util.run_as_user('root')
def render_configs():
    '''Render the configuration templates for all required services'''

    LOG('Rendering squid configuration templates')
    config = hookenv.config()
    context = {}
    # cache_mem() needs to be called before cache_dirs() as it sets
    # unitdata.kv().get('squid-cache-disk').
    context['Cache_Memory_Config'] = cache_mem()
    context['Cache_Storage_Config'] = cache_dirs()
    context['Cache_Peers'] = peer_config()
    context['Cache_Hostname'] = hookenv.unit_private_ip()
    if config['squid_snmp']:
        context[
            'Snmp_Config'
        ] = """acl snmppublic snmp_community public
                                  snmp_port 3401
                                  snmp_access allow snmppublic localhost
                                  snmp_access deny all
                                  snmp_incoming_address 127.0.0.1"""
    else:
        context['Snmp_Config'] = ""

    templating.render('{}/{}.conf'.format(SERVICE, SERVICE), '/etc/{}/{}.conf'.format(SERVICE, SERVICE), context)

    peer_context = {}
    peer_context['Cache_Sibling_IPs'] = '\n'.join(util.peer_IPs())
    templating.render(
        '{}/allowed-networks-src.conf'.format(SERVICE), '/etc/{}/allowed-networks-src.acl'.format(SERVICE), peer_context
    )

    dstdomain_context = {}
    # if 99-ubuntu-repository-cache contains a repository already allowed in
    # the default config, squid-deb-proxy will refuse to start. Hence the
    # following filtering. I don't expect this list to change often (if ever),
    # that is why it's hardcoded.
    default_dstdomains = re.compile(
        '^archive\\.ubuntu\\.com$|'
        '\\.archive\\.ubuntu\\.com$|'
        '^ports\\.ubuntu\\.com$|'
        '^security\\.ubuntu\\.com$|'
        '^ddebs\\.ubuntu\\.com$|'
        '^mirrors\\.ubuntu\\.com$|'
        '^archive\\.canonical\\.com$|'
        '\\.archive\\.canonical\\.com$|'
        '^extras\\.ubuntu\\.com$|'
        '\\.extras\\.ubuntu\\.com$|'
        '^changelogs\\.ubuntu\\.com$'
    )
    if default_dstdomains.match(config['sync-host']):
        dstdomain_context['Sync_Host'] = ''
    else:
        dstdomain_context['Sync_Host'] = config['sync-host']

    templating.render(
        '{}/mirror-dstdomain.acl'.format(SERVICE),
        '/etc/{}/mirror-dstdomain.acl.d/99-ubuntu-repository-cache'.format(SERVICE),
        dstdomain_context,
    )


def update_checks(nrpe_config):
    ''''Update Nagios checks for squid serving archive pool content'''

    path_base = hookenv.config()['path-base']
    nrpe_config.add_check(
        shortname='squid',
        description='squid serving archive pool',
        check_cmd=(
            'check_http --hostname localhost --url "/{}/pool/" ' '--string "main/" --expect \"200\\ OK\"'
        ).format(path_base),
    )
