'''Functions related to charm storage maintenance'''

import glob
import os
import subprocess

from charmhelpers.contrib.storage.linux import utils as Storage
from charmhelpers.core import (
    fstab,
    hookenv,
    host,
    unitdata,
)

LOG = hookenv.log
SERVICE = 'ubuntu-repository-cache'


def get_storage_mounts():
    primary_mount = '/'.join(('/srv', hookenv.service_name()))
    secondary_mounts = []
    if unitdata.kv().get('ephemeral-mounts', None):
        primary_mount = unitdata.kv().get('ephemeral-mounts')[0]
        secondary_mounts = unitdata.kv().get('ephemeral-mounts')[1:]
    return primary_mount, secondary_mounts


def setup():
    '''Set up storage for the charm'''

    config = hookenv.config()
    _setup_ephemeral(config.get('ephemeral-devices', None))

    (primary_mount, secondary_mounts) = get_storage_mounts()

    # Make file system directories
    LOG('Creating filesystem entries for %s' % SERVICE)
    _mkdir(primary_mount, perms=0o775)

    _setup_metadata_path(primary_mount)
    _setup_packages_path(primary_mount, secondary_mounts)


def _setup_metadata_path(rootdir):
    '''Create the directory structure for metadata served by apache'''

    config = hookenv.config()
    apache_root = '/'.join((rootdir, 'apache'))
    unitdata.kv().set('apache-root', apache_root)

    # Apache serves from www/ and metadata is in data/
    _mkdir(apache_root, owner='www-data', group='www-data', perms=0o775)
    _mkdir('/'.join((apache_root, 'www')), owner='www-data', group='www-data', perms=0o775)
    _mkdir('/'.join((apache_root, 'data')), owner='www-data', group='www-data', perms=0o775)
    # XXX support multi-directory path-base
    host.symlink('/'.join((apache_root, 'data', 'ubuntu_active')), '/'.join((apache_root, 'www', config['path-base'])))


def _setup_packages_path(primary_mount, secondary_mounts):
    """Create the squid cache directorys for package caches.

    The config 'squid-disk-caches' is populated with the path to each
    cache directory and the maximum space available, in MB, for that mount.
    In this list the primary device is added last and the maximum space
    calculation reserves room for the OS and metadata mirror.

    The order of the list is important because squid disk caches will
    be sized based on system memory, so some available disk space may
    not be used.  In that case, having the mount used for metadata
    mirrors last would let any additional space be used for that.
    """

    squid_disk_caches = []
    for mnt in secondary_mounts:
        squid_cache = '/'.join((mnt, 'squid'))
        _mkdir(squid_cache, owner='proxy', group='proxy', perms=0o770)
        max_size = _disk_free(squid_cache) * 0.95
        if max_size:
            squid_disk_caches.append([squid_cache, max_size])

    # Add the primary mount last for squid.  This allocates it last
    # and saves additional space for metadata mirrors
    squid_cache = '/'.join((primary_mount, 'squid'))
    _mkdir(squid_cache, owner='proxy', group='proxy', perms=0o770)
    max_size = _disk_free(squid_cache)
    # Reserve 20GB for metadata, extra 5GB headroom, and 4GB for the OS
    # See https://pastebin.ubuntu.com/p/ftCFNMx947/
    reserve = (20 + 5 + 4) * 1024
    if max_size < reserve:
        LOG('Less than {}MB disk free @ {} for mirror.'.format(reserve, primary_mount), hookenv.ERROR)
    else:
        max_size = (max_size - reserve) * 0.75
        squid_disk_caches.append([squid_cache, max_size])

    unitdata.kv().set('squid-disk-caches', squid_disk_caches)


def _setup_ephemeral(devices=None):
    """Formats and mounts ephemeral devices

    Each ephemeral device is formated with ext4 and provided a label
    of the format '<service_name>_<#>'

    Devices are mounted at:
     /srv/<service_name>/ephemeral/<#>

    An fstab entry is added for each device to ensure the device is
    mounted on reboot.
    """

    ephemeral_mounts = []
    if not devices:
        unitdata.kv().set('ephemeral-mounts', [])
        LOG('No ephemeral devices provided')
        return

    base_dir = '/'.join(('/srv', hookenv.service_name()))
    _mkdir(base_dir, perms=0o775)
    ephemeral_dir = '/'.join((base_dir, 'ephemeral'))
    _mkdir(ephemeral_dir, perms=0o775)

    devno = 0
    for dev in devices.split(','):
        dev = dev.strip()
        LOG('Configuring ephemeral device {}.'.format(dev), hookenv.DEBUG)
        if not Storage.is_block_device(dev):
            LOG('%s is not a block device, skipping' % dev, hookenv.ERROR)
            continue

        partitions = glob.glob(dev + '*').remove(dev) or []
        for part in partitions:
            if Storage.is_device_mounted(part):
                LOG('Ephemeral device had a mounted partition %s skipping' % part, hookenv.ERROR)
                continue

        if Storage.is_device_mounted(dev):
            LOG('Unmounting %s' % dev)
            host.umount(dev)

        label = '_'.join(('urc', str(devno)))
        mountpoint = '/'.join((ephemeral_dir, str(devno)))

        Storage.zap_disk(dev)
        subprocess.check_call(['mkfs', '-F', '-t', 'ext4', '-L', label, dev])

        # Remove any existing fstab entry
        ftab = fstab.Fstab()
        entry = ftab.get_entry_by_attr('device', dev)
        if entry:
            ftab.remove_entry(entry)

        # We should always do this anyway, but currently,
        # due to LP#1626318, following a reboot, the ephemeral
        # disk will be rendered immutable and read-only.
        azure_disk_warning = 'ephemeral-disk-warning'
        if host.service_available(azure_disk_warning):
            LOG('Disabling {}'.format(azure_disk_warning))
            host.service_pause(azure_disk_warning)

        # Add to fstab by label, not device name
        _mkdir(mountpoint, perms=0o775)
        devname = '='.join(('LABEL', label))
        ftab.add_entry(fstab.Fstab.Entry(devname, mountpoint, 'ext4', options=None))
        ftab.close()

        subprocess.check_call(['mount', mountpoint])
        os.chmod(mountpoint, 0o775)
        ephemeral_mounts.append(mountpoint)
        devno += 1
    LOG('Ephemeral storage mounted @ {}'.format(ephemeral_mounts), hookenv.DEBUG)
    unitdata.kv().set('ephemeral-mounts', ephemeral_mounts)


def _mkdir(path, owner='root', group='root', perms=0o555):
    """Make directories recursively with a umask=0o000.

    This storage module makes calls to mkdir repeatedly with permissions
    that include group write permission.  This would be masked by the
    default 0o022 umask.   This simple helper function will clear
    the umask prior to calling host.mkdir and then restore it when done.
    """

    prev_umask = os.umask(0o000)
    host.mkdir(path, owner=owner, group=group, perms=perms)
    os.umask(prev_umask)


def _disk_free(path):
    '''Return disk space free, in MB, for the specified path'''

    output = subprocess.check_output(['df', '-BM', path]).decode().split('\n')
    _, _, _, avail, _, _ = output[1].split()
    LOG('disk free for {} is {}MB'.format(path, avail), hookenv.DEBUG)
    return int(avail.rstrip('M'))


def update_checks(nrpe_config):
    '''Update nagios check for ephemeral storage space if present'''

    config = hookenv.config()
    if not config.get('ephemeral-devices'):
        return

    try:
        LOG('Adding check for ephemeral storage {}'.format(unitdata.kv().get('ephemeral-mounts')), hookenv.DEBUG)
    except KeyError:
        LOG('No checks added for ephemeral storage', hookenv.DEBUG)
        return

    nrpe_config.add_check(
        shortname='disks-ephemeral',
        description='disk space on ephemeral volumes',
        check_cmd='check_disk -w 15 -c 5 -r ' '^/srv/ubuntu-repository-cache/ephemeral/[0-9].*$',
    )
