import itertools
import os
import shutil
import subprocess
import tempfile
import unittest

from collections import OrderedDict, namedtuple
from nose.tools import assert_raises

from .. import metadata_sync

RELEASE_FILE_CONTENT = """
Origin: Ubuntu
Label: Ubuntu
Blah: blah
MD5Sum:
 5997256e5f20523338c22a9caefd7c95        435597291 md5-something
SHA1:
 dc8c2798c3d6707e0b313fa8a9b50d0682310217        435597291 something
 a41cb792e56b512948b9dd35d3c974497d84f1ea         47334442 something-else
"""

RELEASE_FILE_WITH_SHA256 = (
    RELEASE_FILE_CONTENT
    + """
SHA256:
 36193e81e070279baaf8adce44305b02c3fd029f6c0f1497d9bd044ab2a633aa 435597291 foo
 e653b6621b1d77144d86421927ae7f9466f9a6b7012fa4b9a51cdf68b333a23a  47334442 bar
"""
)


EXPECTED_FILES_DICT = OrderedDict(
    [
        ('foo', {'SHA256': '36193e81e070279baaf8adce44305b02c3fd029f6c0f1497d9bd' '044ab2a633aa', 'size': '435597291'}),
        ('bar', {'SHA256': 'e653b6621b1d77144d86421927ae7f9466f9a6b7012fa4b9a51c' 'df68b333a23a', 'size': '47334442'}),
    ]
)


FakeException = namedtuple('FakeException', 'returncode')


class InvocationRecorder(object):
    """A simple object that records every invocation.

    All invocations of an  instance's __call__() are recorded to the
    @ivar{invocations}.
    """

    def __init__(self):
        self.invocations = []

    def __call__(self, *args, **kwargs):
        self.invocations.append([args, kwargs])


class MirrorTestCase(unittest.TestCase):
    def test_rsync_sends_rsync_command(self):
        """
        The rsync function executes a system rsync shell with the passed in
        parameters.
        """

        def fake_check_output(cmd, stderr):
            self.assertEqual(stderr, subprocess.STDOUT)
            self.assertEqual(["/usr/bin/rsync", "-r", "--foo", "--bar", "source", "dest"], cmd)
            return b"called "  # Note: real check_output returns bytes.

        log = InvocationRecorder()

        options = ("--foo", "--bar")
        result = metadata_sync.rsync("source", "dest", options, log, check_output=fake_check_output)

        self.assertEqual("called", result)  # "rsync" decodes and strips
        self.assertEqual([[("/usr/bin/rsync -r --foo --bar source dest",), {}]], log.invocations)

    def test_parse_release_file_no_sha256(self):
        """
        Parsing a minimal release file returns the expected dictionary with no
        mapping of hashes to files.
        """
        result = None
        temp_dir = tempfile.mkdtemp()
        try:
            release = os.path.join(temp_dir, "Release")
            with open(release, "w") as temp:
                temp.write(RELEASE_FILE_CONTENT)
                temp.flush()
            result = metadata_sync._parse_release_file(temp_dir)
            expected = {
                'files': OrderedDict(),
                'name': os.path.basename(temp_dir),
                'mtime': os.stat(temp_dir + "/Release").st_mtime,
            }
            self.assertEqual(expected, result)
        finally:
            shutil.rmtree(temp_dir)

    def test_parse_release_file_with_sha256(self):
        """
        Parsing a minimal release file with sha256 hashes returns a mapping
        of hashes to files.
        """
        result = None
        temp_dir = tempfile.mkdtemp()
        try:
            release = os.path.join(temp_dir, "Release")
            with open(release, "w") as temp:
                temp.write(RELEASE_FILE_WITH_SHA256)
                temp.flush()
            result = metadata_sync._parse_release_file(temp_dir)
            expected = {
                'name': os.path.basename(temp_dir.strip("/")),
                'mtime': os.stat(temp_dir + "/Release").st_mtime,
                'files': EXPECTED_FILES_DICT,
            }
            self.assertEqual(expected, result)
        finally:
            shutil.rmtree(temp_dir)

    def test_rsync_should_retry_fail(self):
        """Rsync exit codes 1 to 4 are real failures."""
        for i in range(1, 5):
            exe = FakeException(returncode=i)
            self.assertTrue(metadata_sync.rsync_should_giveup(exe))

    def test_rsync_should_retry_do_retry(self):
        """Retry for whitelisted exit codes."""
        retriables = frozenset((5, 6, 10, 11, 12, 13, 14, 20, 21, 22, 23, 24, 25, 30, 35))

        for i in retriables:
            exe = FakeException(returncode=i)
            self.assertFalse(metadata_sync.rsync_should_giveup(exe))

    def test_rsync_should_retry_unknown(self):
        """An unknown rsync return value is not retried."""
        exe = FakeException(returncode=7)
        self.assertTrue(metadata_sync.rsync_should_giveup(exe))

    def test_main_all_variables(self):
        """
        The main function fails should an environment variable be missing.
        """

        def noop(*args, **kwargs):
            return "Called"

        env = {
            "SYNC_HOST": "foo.com",
            "MIRROR_SERIES": "xenial",
            "APACHE_ROOT": "/var/www/",
            "LOCAL_UNIT": "unit/0",
            "UNIT_PATH": "something",
        }
        result = metadata_sync.main(env, InvocationRecorder(), mirror_archive=noop, check_call=noop)
        self.assertEqual("Called", result)

    def test_main_calls_juju_run(self):
        """
        The main() function calls juju run once it determined the meta_ver from
        syncing with upstream.
        """
        env = {
            "SYNC_HOST": "foo.com",
            "MIRROR_SERIES": "xenial",
            "APACHE_ROOT": "/var/www/",
            "LOCAL_UNIT": "unit/0",
            "UNIT_PATH": "something",
        }

        def fake_mirror_archive(*args, **kwargs):
            return "some_meta_var"

        fake_check_call = InvocationRecorder()

        result = metadata_sync.main(
            env, InvocationRecorder(), mirror_archive=fake_mirror_archive, check_call=fake_check_call
        )

        self.assertEqual("some_meta_var", result)

        expected = [
            [
                (
                    [
                        'juju-run',
                        'unit/0',
                        '/var/lib/juju/agents/something/charm/bin/charm-env '
                        '/var/lib/juju/agents/something/charm/hooks/'
                        'ubuntu-repository-cache-sync some_meta_var',
                    ],
                ),
                {},
            ]
        ]

        self.assertEqual(expected, fake_check_call.invocations)


def _filter_dict(d, keys):
    """Return a new dictionary containing only `keys` from `d`."""
    return dict((k, v) for k, v in d.items() if k in keys)


def test_main_missing_environment_variable():
    # The main function fails should an environment variable be missing.
    full_dict = {
        "SYNC_HOST": "foo.com",
        "APACHE_ROOT": "/var/www/",
        "LOCAL_UNIT": "unit/0",
        "UNIT_PATH": "something",
    }
    key_combinations = itertools.chain(
        *[itertools.combinations(full_dict.keys(), n) for n in range(len(full_dict.keys()))]
    )

    def do_test(env):
        assert_raises(KeyError, metadata_sync.main, env, InvocationRecorder())

    for key_set in key_combinations:
        env = _filter_dict(full_dict, key_set)
        yield do_test, env
