import os
import shutil
import tempfile
import unittest
from unittest import mock

from charmhelpers.core import unitdata

from .. import squid


class SquidTestCase(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.tmpdir = tempfile.mkdtemp(prefix='charm-unittests-')
        self.addCleanup(shutil.rmtree, self.tmpdir)

        patcher = mock.patch('charmhelpers.core.hookenv.config')
        self.mock_config = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch('charmhelpers.core.hookenv.log')
        self.mock_log = patcher.start()
        self.addCleanup(patcher.stop)

        os.environ['UNIT_STATE_DB'] = os.path.join(self.tmpdir, '.unit-state.db')

    def test_squid_cache_size(self):
        unitdata.kv().set('squid-cache-disk', 5000)
        unitdata.kv().set('squid-disk-caches', [('/srv', 2000)])

        self.mock_config.return_value = {
            'squid-cache-disk': 20 * 1024,
            'squid-disk-caches': [['/srv', 2000]],
            'cache-storage-size': 0,
        }
        result = squid.cache_dirs()
        self.assertEqual(result, 'cache_dir aufs /srv 2000 16 256\n')

        self.mock_config.return_value = {
            'squid-cache-disk': 20 * 1024,
            'squid-disk-caches': [['/srv', 2000]],
            'cache-storage-size': 1000,
        }
        result = squid.cache_dirs()
        self.assertEqual(result, 'cache_dir aufs /srv 1000 16 256\n')

    def test_squid_cache_mem(self):
        # Lots of system RAM, 14G per Azure D3_v2. 50GB cache, auto calculate memory usage.
        self.mock_config.return_value = {
            'cache-storage-size': 50000,
            'cache-memory-size': 0,
        }
        result = squid.cache_mem(13997)
        self.assertEqual(result, 'cache_mem 3122 MB')

        # 7GB per Azure D2_v2. 50GB cache, auto calculate memory usage.
        self.mock_config.return_value = {
            'cache-storage-size': 50000,
            'cache-memory-size': 0,
        }
        result = squid.cache_mem(6954)
        self.assertEqual(result, 'cache_mem 1009 MB')

        # 7GB per Azure D2_v2. 50GB cache, fixed 1000 MB.
        self.mock_config.return_value = {
            'cache-storage-size': 50000,
            'cache-memory-size': 1000,
        }
        result = squid.cache_mem(6954)
        self.assertEqual(result, 'cache_mem 1000 MB')

        # 7GB per Azure D2_v2. Disk with 50GB, auto calculate memory usage.
        unitdata.kv().set('squid-disk-caches', [('/srv', 50000)])
        self.mock_config.return_value = {
            'cache-storage-size': 0,
            'cache-memory-size': 0,
        }
        result = squid.cache_mem(6954)
        self.assertEqual(result, 'cache_mem 1009 MB')
