'''Miscellaneous helpers for the ubuntu-repository-cache charm '''

import functools
import glob
import os
import pwd
import subprocess

from charmhelpers.contrib.hahelpers import cluster
from charmhelpers.core import (
    hookenv,
    host,
    unitdata,
)

LOG = hookenv.log
SERVICE = 'ubuntu-repository-cache'

# Log file and directory for the repository sync cron job
REPO_SYNC_DIR = '/var/log/repository-sync'
REPO_SYNC_CRON_LOG = os.path.join(REPO_SYNC_DIR, 'repository-sync-cron.log')

# A file we'll touch when the metadata sync is running. Another run of the sync
# will noop while this file is present.
SEMAPHORE_FILE = '/tmp/metadata-sync-running'


class SemaphoreExistsError(Exception):
    '''An error raised when the desired semaphore name already exists.'''


class FileSemaphore:
    """A context manager that creates and deletes semaphores at the given path.

    It will raise SemaphoreExistsError if the "lock" can't be acquired, and
    will cleanup after itself on exit."""

    def __init__(self, filename):
        self.filename = filename

    def __enter__(self):
        if os.path.exists(self.filename):
            raise SemaphoreExistsError("File '{}' already exists.".format(self.filename))
        touch(self.filename)

    def __exit__(self, *args):
        os.remove(self.filename)


def touch(filename):
    """Make an empty file at the passed in filename.

    Its modified access time is set to "now"."""
    with open(filename, 'a'):
        os.utime(filename, None)


def in_failover():
    '''Return whether the unit is in the failover state'''

    return unitdata.kv().get('failover', False)


def set_failover(failover=True):
    '''Set the failover state for the unit'''

    unitdata.kv().set('failover', failover)


def unset_failover():
    '''Clear the failover state for the unit'''

    set_failover(failover=False)


def get_failover_host():
    """Find the private address of the failover host.

    The next highest numbered unit in the cluster relationship is
    choosen as the host for failover.  Its private address is returned.
    """

    if not hookenv.is_relation_made('cluster'):
        return None

    # The leader does not failover to another host
    if cluster.is_elected_leader(None):
        return None

    rel_id = hookenv.relation_ids(reltype='cluster')[0]
    cluster_units = hookenv.related_units(relid=rel_id)
    if not cluster_units:
        return None

    cluster_ids = [int(x.split('/')[1]) for x in cluster_units]
    local_id = int(hookenv.local_unit().split('/')[1])

    # Filter the list to find all units with a higher unit id, will
    # wrap to lowest unit number
    cluster_ids = [x for x in cluster_ids if x > local_id] or [x for x in cluster_ids if x < local_id]
    if not cluster_ids:
        return None

    failover_unit = '%s/%s' % (hookenv.service_name(), cluster_ids[0])
    return hookenv.relation_get(attribute='private-address', unit=failover_unit, rid=rel_id)


def restore_saved_id():
    '''Restore the effective uid/gid from the saved uid/gid'''

    os.seteuid(os.getresuid()[2])
    os.setegid(os.getresgid()[2])


# Submit MP to include this in charmhelpers.core.host
def run_as_user(user_name):
    """A decorator to run a function as a particular user.

    This works as long as the saved uid (see os.getresuid()[2])
    has permission to change the effective to that of the
    desired uid/gid of the provided user name.

    The $HOME environment variable is also set to the particular user.
    """

    def decorator(function):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            # Remember the current effective uid/gid
            orig_id = os.geteuid(), os.getegid()
            orig_home = os.environ.get('HOME')

            # Change to saved uid/gid so that wrapped functions can nest
            # This assumes that the saved id has
            restore_saved_id()

            # Set effective uid/gid for specified user
            pw_record = pwd.getpwnam(user_name)
            os.setegid(pw_record.pw_gid)
            os.seteuid(pw_record.pw_uid)
            os.environ['HOME'] = pw_record.pw_dir

            result = function(*args, **kwargs)

            # Restore effective uid/gid and home
            restore_saved_id()
            os.setegid(orig_id[1])
            os.seteuid(orig_id[0])
            if orig_home:
                os.environ['HOME'] = orig_home
            elif 'HOME' in os.environ:
                del os.environ['HOME']

            return result

        return wrapper

    return decorator


def system_total_mem():
    '''Return the total system memory in megabytes'''

    output = subprocess.check_output(['free', '-m']).decode().split('\n')
    return int(output[1].split()[1])


def change_aptsources_url(url='http://archive.ubuntu.com/ubuntu'):
    """Change the URL for the current apt sources

    This affects the URL for all sources in /etc/apt/sources.list
    and /etc/apt/sources.list.d/ so care must be taken if the charm
    has added any custom source URLs as it will alter them as well.
    """
    import aptsources.sourceslist
    import aptsources.distro

    sourceslist = aptsources.sourceslist.SourcesList()
    sourceslist.refresh()
    if not glob.glob('/etc/apt/sources.list.%s' % SERVICE):
        sourceslist.backup(backup_ext='.'.join(['', SERVICE, 'save']))
    distro = aptsources.distro.get_distro()
    distro.get_sources(sourceslist)
    distro.change_server(url)
    sourceslist.save()


def peer_IPs(reltype='cluster'):
    '''Return a list of IPs for the peers in the specified relation type'''

    ips = []
    for ident in hookenv.relation_ids(reltype=reltype):
        for unit in hookenv.related_units(relid=ident):
            ips.append(hookenv.relation_get(attribute='private-address', unit=unit, rid=ident))
    return ips


def get_active_metaver():
    """Get the metadata version of the active metadata

    Returns version or None
    """

    apache_data = '/'.join((unitdata.kv().get('apache-root'), 'data'))
    path = os.path.realpath('/'.join((apache_data, 'ubuntu_active')))
    meta_ver = None
    if os.path.exists(path):
        meta_ver = os.path.basename(path)
    return meta_ver


def update_active_symlink(meta_version, apache_root=None):
    """Update the symlink to the active metadata to point to the
    metadata for meta_version."""

    if apache_root:
        root = apache_root
    else:
        root = unitdata.kv().get('apache-root')
    apache_data = '/'.join((root, 'data'))
    meta_dir = '/'.join((apache_data, meta_version))
    next_link = '/'.join((apache_data, 'ubuntu_next'))
    active_link = '/'.join((apache_data, 'ubuntu_active'))

    if os.path.exists(next_link):
        os.remove(next_link)
    host.symlink(meta_dir, next_link)
    subprocess.check_call(['mv', '-T', next_link, active_link])


def remote_path_exists(user, host_ip, path, localuser='www-sync'):
    """Use ssh to determine if a remote path exists.

    Remotely execute the stat(1) command on the remote host."""

    cmd = ('sudo', '-u', localuser, 'ssh', '{}@{}'.format(user, host_ip), 'stat {}'.format(path))
    try:
        subprocess.check_call(cmd)
    except subprocess.CalledProcessError as excp:
        # ssh error
        if excp.returncode == 255:
            raise
        else:  # path does not exist
            return False
    return True
