# Utility methods and constants for functional tests

import amulet
import logging
import os
import requests

# Fetch the charm location from CHARM_DIR because bundletester relocates the
# tests when executing
CHARM_DIR = os.environ.get('CHARM_DIR')
MIRROR_SERIES = os.environ.get('MIRROR_SERIES', '')
SERVICE = 'ubuntu-repository-cache'
TIMEOUT = (40 + 60) * 60

CRON_FILE = '/etc/cron.d/ubuntu-repository-cache_rsync'
LOG_DIR = '/var/log/repository-sync'
LOG_NAME = 'repository-sync-cron.log'
LOG_FILE = os.path.join(LOG_DIR, LOG_NAME)

# Hijack the logger from charm test
LOGGER = logging.getLogger('juju-test')


def setup_logging():
    ft = logging.Formatter('%(asctime)s %(name)s %(levelname)-8s: %(message)s')
    sh = logging.StreamHandler()
    sh.setFormatter(ft)
    LOGGER.addHandler(sh)


def log(msg):
    LOGGER.warn(''.join(('test: ', msg)))


def unit_name(unit):
    """
    Returns the juju unit name for the given amulet unit object.
    """
    return unit.info['unit_name']


def send_script(unit, path, script):
    """
    Write out the provided script contents to the specified path and unit.

    Writes one line out at a time using ssh since there is no other convenient
    amulet method to do so. Works fine for short scripts.
    """
    unit.ssh('sudo rm -f {}'.format(path))
    for line in script.splitlines():
        command = 'sudo echo \'{}\' >> {}'.format(line, path)
        unit.ssh(command)
    unit.ssh('sudo chmod +x {}'.format(path))


def execute(unit, command):
    """
    Executes a command on the given unit via a wrapper script.

    The amulet ssh method provides excessively verbose output on errors.
    This wrapper gives the test more control over the output and result.
    """
    log('INFO: Executing: {}'.format(command))
    unit.ssh('sudo rm -f /tmp/command-runner.result')
    unit.ssh('sudo rm -f /tmp/command-runner.log')
    runner_template = '\n'.join(
        [
            '#!/bin/bash',
            '{} > /tmp/command-runner.log 2>&1',
            'echo $? > /tmp/command-runner.result',
        ]
    )
    runner_script_path = '/tmp/command-runner.sh'
    runner_script = runner_template.format(command)
    send_script(unit, runner_script_path, runner_script)
    unit.ssh('sudo {}'.format(runner_script_path))
    result = unit.file_contents('/tmp/command-runner.result').strip()
    output = unit.file_contents('/tmp/command-runner.log')
    try:
        return (int(result), output)
    except ValueError:
        msg = 'Command execution failed on {}:\n' '\texit code: {}\n\toutput: {}'.format(
            unit_name(unit), result, output
        )
        amulet.raise_status(amulet.FAIL, msg=msg)


def disable_cron(unit):
    """
    Disables the cron service on the given unit.
    """
    command = 'sudo systemctl stop cron.service'
    (result, output) = execute(unit, command)
    if result:
        msg = 'Failed to disable cron on {}:\n' '\texit code: {}\noutput: {}'.format(unit_name(unit), result, output)
        amulet.raise_status(amulet.FAIL, msg=msg)


def rotate_cron_sync_log(unit):
    """
    Rotates the LOG_FILE via logrotate.

    The end result is an empty LOG_FILE.
    """
    command = 'sudo logrotate -f /etc/logrotate.d/apache2'
    (result, output) = execute(unit, command)
    if result:
        # logrotate can return non-zero even if the log file we care
        # about was rotated
        log(
            'INFO: Rotating logs on {} return non-zero result (may be ok):\n'
            '\texit code: {}\noutput: {}'.format(unit_name(unit), result, output)
        )


def attempt_rsync(unit):
    """
    Performs an archive sync as specified by the cron file.

    This extracts the user, environment variables and command from the unit's
    cron file and assembles it all into a command to be executed via juju ssh.

    It would be preferred to let the cron run, but this allows for easier
    extraction of any errors.
    """
    contents = unit.file_contents(CRON_FILE)
    lines = contents.splitlines()

    # Extract the environment variables
    env_vars = ' '.join([line for line in lines if '=' in line])

    # Extract the line with the command to execute and parse out the parts we
    # need to run it via ssh
    cron_line = lines[-1].split()
    cron_user = cron_line[5]
    cron_command = ' '.join(cron_line[6:])
    command = 'sudo -Hu {} {} {}'.format(cron_user, env_vars, cron_command)
    (result, output) = execute(unit, command)
    if result:
        msg = 'Failed to execute repository sync command on {}\n' '\tresult: {}\n\toutput: {}'.format(
            unit_name(unit), result, output
        )
        amulet.raise_status(amulet.FAIL, msg=msg)


def check_message_count(contents, message, count):
    """
    Check for the requested message in the content.
    """
    found = len([line for line in contents.splitlines() if message in line])
    if count == found:
        log('PASS: "{}" found {} times'.format(message, count))
    else:
        msg = 'FAIL: "{}" found {} times, expected {}'.format(message, found, count)
        amulet.raise_status(amulet.FAIL, msg=msg)


def test_url(url, waitsec=30):
    """Request a url and return the contents.

    Return:
        response on success,
        None on timeout,
        Raise exception for all other failures
    """

    try:
        page = requests.get(url, timeout=waitsec)
    except requests.exceptions.Timeout:
        return None
    except Exception:
        raise
    return page


def check_metadata_is_active(unit):
    """
    Checks that the synced metadata is available over http.
    """
    page = test_url('http://{}/ubuntu'.format(unit.info['public-address']))
    if page is None:
        log('FAIL: Timeout checking metadata on {}'.format(unit_name(unit)))
        return
    page.raise_for_status()
    log('PASS: Metadata is active on {}'.format(unit_name(unit)))


def check_pool_data_is_active(unit):
    """
    Checks that the squid cached pool data is available over http.
    """
    page = test_url('http://{}/ubuntu/pool'.format(unit.info['public-address']))
    if page is None:
        log('FAIL: Timeout checking pool data on {}'.format(unit_name(unit)))
        return
    page.raise_for_status()
    log('PASS: Pool data is active on {}'.format(unit_name(unit)))
